import decimal

from django import forms
from django.utils.translation import gettext_lazy as _

from djangocms_frontend.models import FrontendUIItem

from ...helpers import coerce_decimal
from .entry_model import FormEntry  # noqa


class Form(FrontendUIItem):
    class Meta:
        proxy = True
        verbose_name = _("Form")

    def get_short_description(self):
        name = self.config.get("form_name", None)
        if name:
            return f"({name})"
        return "<unnamed>"


class FormFieldMixin:
    def get_short_description(self):
        label = self.config.get("field_label", "")
        return f"{label} ({self.config.get('field_name')})"


class CharField(FormFieldMixin, FrontendUIItem):
    class Meta:
        proxy = True
        verbose_name = _("Character field")

    def get_form_field(self):
        return self.field_name, forms.CharField(
            label=self.config.get("field_label", ""),
            required=self.config.get("field_required", False),
            widget=forms.TextInput(
                attrs=dict(placeholder=self.config.get("field_placeholder", ""))
            ),
        )


class EmailField(FormFieldMixin, FrontendUIItem):
    class Meta:
        proxy = True
        verbose_name = _("Email field")

    def get_form_field(self):
        return self.field_name, forms.EmailField(
            label=self.config.get("field_label", ""),
            required=self.config.get("field_required", False),
        )


class UrlField(FormFieldMixin, FrontendUIItem):
    class Meta:
        proxy = True
        verbose_name = _("URL field")

    def get_form_field(self):
        return self.field_name, forms.URLField(
            label=self.config.get("field_label", ""),
            required=self.config.get("field_required", False),
        )


class DecimalField(FormFieldMixin, FrontendUIItem):
    class Meta:
        proxy = True
        verbose_name = _("Decimal field")

    class NumberInput(forms.NumberInput):
        def __init__(self, **kwargs):
            self.decimal_places = kwargs.pop("decimal_places", None)
            super().__init__(**kwargs)

        def format_value(self, value):
            if value is None:
                return ""
            value = str(value)
            if "." in value:
                l, r = value.rsplit(".", 1)
            else:
                l, r = value, ""
            if self.decimal_places == 0:
                return l
            r = (r + self.decimal_places * "0")[: self.decimal_places]
            return super().format_value(".".join((l, r)))

    class StrDecimalField(forms.DecimalField):
        def clean(self, value):
            value = super().clean(value)
            if isinstance(value, decimal.Decimal):
                value = str(value)
            return value

    def get_form_field(self):
        return self.field_name, DecimalField.StrDecimalField(
            label=self.config.get("field_label", ""),
            required=self.config.get("field_required", False),
            min_value=coerce_decimal(self.config.get("min_value", None)),
            max_value=coerce_decimal(self.config.get("max_value", None)),
            decimal_places=self.config.get("decimal_places", None),
            widget=DecimalField.NumberInput(
                decimal_places=self.config.get("decimal_places", None)
            ),
        )


class IntegerField(FormFieldMixin, FrontendUIItem):
    class Meta:
        proxy = True
        verbose_name = _("Integer field")

    def get_form_field(self):
        return self.field_name, forms.IntegerField(
            label=self.config.get("field_label", ""),
            required=self.config.get("field_required", False),
        )


class TextareaField(FormFieldMixin, FrontendUIItem):
    class Meta:
        proxy = True
        verbose_name = _("Text field")

    def get_form_field(self):
        return self.field_name, forms.CharField(
            label=self.config.get("field_label", ""),
            required=self.config.get("field_required", False),
            widget=forms.Textarea(
                attrs=dict(
                    rows=self.config.get("field_rows", 10),
                    placeholder=self.config.get("field_placeholder", ""),
                    style="height: inherit;",
                )
            ),
        )


class DateField(FormFieldMixin, FrontendUIItem):
    class Meta:
        proxy = True
        verbose_name = _("Date field")

    class DateInput(forms.DateInput):
        input_type = "date"

    def get_form_field(self):
        return self.field_name, forms.DateField(
            label=self.config.get("field_label", ""),
            required=self.config.get("field_required", False),
            widget=DateField.DateInput(),
        )


class DateTimeField(FormFieldMixin, FrontendUIItem):
    class Meta:
        proxy = True
        verbose_name = _("Date field")

    class DateTimeField(forms.DateTimeField):
        def prepare_value(self, value):
            if isinstance(value, str):
                from django.utils import dateparse

                return super().prepare_value(dateparse.parse_datetime(value))
            return super().prepare_value(value)

    class DateTimeInput(forms.DateTimeInput):
        input_type = "datetime-local"

    def get_form_field(self):
        return self.field_name, DateTimeField.DateTimeField(
            label=self.config.get("field_label", ""),
            required=self.config.get("field_required", False),
            widget=DateTimeField.DateTimeInput(),
        )


class TimeField(FormFieldMixin, FrontendUIItem):
    class Meta:
        proxy = True
        verbose_name = _("Date field")

    class TimeInput(forms.TimeInput):
        input_type = "time"

    def get_form_field(self):
        return self.field_name, forms.TimeField(
            label=self.config.get("field_label", ""),
            required=self.config.get("field_required", False),
            widget=TimeField.TimeInput(),
        )


class Select(FormFieldMixin, FrontendUIItem):
    class Meta:
        proxy = True
        verbose_name = _("Select")

    _choices = None
    no_selection = [("", _("No selection"))]

    def get_choices(self):
        if self._choices is None:
            descendants = self.get_children().order_by("position")
            self._choices = []
            for child in descendants:
                instance = child.djangocms_frontend_frontenduiitem
                self._choices.append(
                    (instance.config["value"], instance.config["verbose"])
                )
        return self._choices

    def get_form_field(self):
        multiple_choice = self.config.get("field_select", "") in (
            "multiselect",
            "checkbox",
        )
        field = forms.MultipleChoiceField if multiple_choice else forms.ChoiceField
        required = self.config.get("field_required", False)
        choices = self.get_choices()
        if not required and not multiple_choice:
            choices = self.no_selection + choices
        widget_choice = self.config.get("field_select", "")
        if widget_choice == "select":
            widget = forms.Select()
        elif widget_choice == "radio":
            widget = forms.RadioSelect()
        elif widget_choice == "multiselect":
            widget = forms.SelectMultiple(attrs=dict(style="min-height: 6em;"))
        else:
            widget = forms.CheckboxSelectMultiple()

        return self.field_name, field(
            label=self.config.get("field_label", ""),
            required=required,
            choices=choices,
            widget=widget,
        )


class Choice(FormFieldMixin, FrontendUIItem):
    class Meta:
        proxy = True
        verbose_name = _("Choice")

    def get_short_description(self):
        return f'{self.config.get("verbose", "-")} ("{self.config.get("value", "")}")'


class SwitchInput(forms.CheckboxInput):
    pass


class BooleanField(FormFieldMixin, FrontendUIItem):
    class Meta:
        proxy = True
        verbose_name = _("Boolean field")

    def get_form_field(self):
        return self.field_name, forms.BooleanField(
            label=self.config.get("field_label", ""),
            required=self.config.get("field_required", False),
            widget=SwitchInput()
            if self.config.get("field_as_switch", False)
            else forms.CheckboxInput(),
        )


class SubmitButton(FormFieldMixin, FrontendUIItem):
    class Meta:
        proxy = True
        verbose_name = _("Submit button")


try:
    """Soft dependency on django-captcha for reCaptchaField"""

    from captcha.fields import ReCaptchaField  # NOQA
    from captcha.widgets import (  # NOQA
        ReCaptchaV2Checkbox,
        ReCaptchaV2Invisible,
        ReCaptchaV3,
    )
except ImportError:
    ReCaptchaV2Invisible = forms.HiddenInput  # NOQA
    ReCaptchaV2Checkbox = forms.HiddenInput  # NOQA
    ReCaptchaV3 = forms.HiddenInput  # NOQA

    class ReCaptchaField:  # NOQA
        def __init__(self, *args, **kwargs):
            pass



from django import forms
from django.conf import settings
from django.utils.translation import gettext_lazy as _

from djangocms_frontend.helpers import coerce_decimal

try:
    """Soft dependency on django-captcha for reCaptchaField"""

    from captcha.fields import ReCaptchaField  # NOQA
    from captcha.widgets import (  # NOQA
        ReCaptchaV2Checkbox,
        ReCaptchaV2Invisible,
        ReCaptchaV3,
    )

    installed = True

except ImportError:
    ReCaptchaV2Invisible = forms.HiddenInput  # NOQA
    ReCaptchaV2Checkbox = forms.HiddenInput  # NOQA
    ReCaptchaV3 = forms.HiddenInput  # NOQA
    installed = False

    class ReCaptchaField:  # NOQA
        def __init__(self, *args, **kwargs):
            pass


WIDGETS = {
    "v2-checkbox": ReCaptchaV2Checkbox,
    "v2-invisible": ReCaptchaV2Invisible,
}


RECAPTCHA_CHOICES = (
    ("v2-checkbox", _("v2 checkbox")),
    ("v2-invisible", _("v2 invisible")),
)


def get_recaptcha_field(config):
    widget_params = {
        "attrs": {
            key: value
            for key, value in config.get("captcha_config", {}).items()
            if key.startswith("data-")
        },
        "api_params": {
            key: value
            for key, value in config.get("captcha_config", {}).items()
            if not key.startswith("data-")
        },
    }
    widget_params["attrs"]["no_field_sep"] = True
    if config.get("captcha_widget", "") == "v3":
        widget_params["attrs"]["required_score"] = coerce_decimal(
            config.get("captcha_requirement", 0.5)
        )
    field = ReCaptchaField(
        widget=WIDGETS[config.get("captcha_widget", "v2-invisible")](
            **widget_params,
        ),
    )
    return field


keys_available = installed and (
    hasattr(settings, "RECAPTCHA_PUBLIC_KEY")
    and hasattr(settings, "RECAPTCHA_PRIVATE_KEY")
)

field_name = "recaptcha_field"
RECAPTCHA_PUBLIC_KEY = getattr(settings, "RECAPTCHA_PUBLIC_KEY", "")



from djangocms_frontend import settings

if settings.framework == "bootstrap5":
    from .bootstrap5.sizing import SizingFormMixin, SizingMixin
else:

    class SizingMixin:
        pass

    class SizingFormMixin:
        pass



from importlib import import_module

from django.apps import AppConfig
from django.conf import settings
from django.urls import NoReverseMatch, clear_url_caches, include, path, reverse
from django.utils.translation import gettext_lazy as _


class FormsConfig(AppConfig):
    name = "djangocms_frontend.contrib.frontend_forms"
    verbose_name = _("Forms")

    def ready(self):
        """Install the URLs"""

        from django.core.checks import Error, Warning, register

        @register()
        def deprecation_check(app_configs, **kwargs):
            if "djangocms_form_builder" in settings.INSTALLED_APPS:
                return [
                    Error(
                        "djangocms_frontend.contrib.frontend_forms and djangocms_form_builder are installed.",
                        obj=self,
                        hint="Remove djangocms_frontend.contrib.frontend_forms",
                        id="djangocms_frontend.E001",
                    )
                ]
            return [
                Warning(
                    "djangocms_frontend.contrib.frontend_forms will be removed in version 1.0",
                    obj=self,
                    hint="Use djangocms-form-builder instead.",
                    id="djangocms_frontend.W001",
                )
            ]

        try:
            reverse("dcf_forms:ajax_form")
        except NoReverseMatch:  # Not installed yet
            urlconf_module = import_module(settings.ROOT_URLCONF)
            urlconf_module.urlpatterns = [
                path(
                    "@dcf-frontend_forms/",
                    include(
                        "djangocms_frontend.contrib.frontend_forms.urls",
                        namespace="dcf_forms",
                    ),
                )
            ] + urlconf_module.urlpatterns
            clear_url_caches()



import inspect

from cms.api import add_plugin
from cms.test_utils.testcases import CMSTestCase

from djangocms_frontend.contrib.frontend_forms import cms_plugins
from djangocms_frontend.contrib.frontend_forms.cms_plugins.form_plugins import (
    FormElementPlugin,
)

from ..fixtures import TestFixture


class FormEditorTestCase(TestFixture, CMSTestCase):
    def test_form_editor(self):
        form = add_plugin(
            placeholder=self.placeholder,
            plugin_type=cms_plugins.FormPlugin.__name__,
            language=self.language,
            config=dict(
                form_selection="",
                form_name="my-test-form",
            ),
        )
        form.initialize_from_form()

        for item, cls in cms_plugins.__dict__.items():
            if inspect.isclass(cls) and issubclass(cls, FormElementPlugin):
                field = add_plugin(
                    placeholder=self.placeholder,
                    plugin_type=cls.__name__,
                    target=form,
                    language=self.language,
                    config=dict(
                        field_name="field_" + item,
                    ),
                )
                field.initialize_from_form()

        self.publish(self.page, self.language)

        with self.login_user_context(self.superuser):
            response = self.client.get(self.request_url)
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, 'action="/@dcf-frontend_forms/1"')
        self.assertContains(response, '<input type="hidden" name="csrfmiddlewaretoken"')
        for item, cls in cms_plugins.__dict__.items():
            if inspect.isclass(cls) and issubclass(cls, FormElementPlugin):
                self.assertContains(response, f'name="field_{item}"')



# Generated by Django 3.2.12 on 2022-03-21 08:15

import django.core.serializers.json
import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models

import djangocms_frontend.contrib.frontend_forms.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("djangocms_frontend", "0001_initial"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="CharField",
            fields=[],
            options={
                "verbose_name": "Character field",
                "proxy": True,
                "indexes": [],
                "constraints": [],
            },
            bases=(
                djangocms_frontend.contrib.frontend_forms.models.FormFieldMixin,
                "djangocms_frontend.frontenduiitem",
            ),
        ),
        migrations.CreateModel(
            name="Form",
            fields=[],
            options={
                "verbose_name": "Form",
                "proxy": True,
                "indexes": [],
                "constraints": [],
            },
            bases=("djangocms_frontend.frontenduiitem",),
        ),
        migrations.CreateModel(
            name="BooleanField",
            fields=[],
            options={
                "verbose_name": "Boolean field",
                "proxy": True,
                "indexes": [],
                "constraints": [],
            },
            bases=(
                djangocms_frontend.contrib.frontend_forms.models.FormFieldMixin,
                "djangocms_frontend.frontenduiitem",
            ),
        ),
        migrations.CreateModel(
            name="Select",
            fields=[],
            options={
                "verbose_name": "Select",
                "proxy": True,
                "indexes": [],
                "constraints": [],
            },
            bases=(
                djangocms_frontend.contrib.frontend_forms.models.FormFieldMixin,
                "djangocms_frontend.frontenduiitem",
            ),
        ),
        migrations.CreateModel(
            name="SubmitButton",
            fields=[],
            options={
                "verbose_name": "Submit button",
                "proxy": True,
                "indexes": [],
                "constraints": [],
            },
            bases=(
                djangocms_frontend.contrib.frontend_forms.models.FormFieldMixin,
                "djangocms_frontend.frontenduiitem",
            ),
        ),
        migrations.CreateModel(
            name="TextareaField",
            fields=[],
            options={
                "verbose_name": "Text field",
                "proxy": True,
                "indexes": [],
                "constraints": [],
            },
            bases=(
                djangocms_frontend.contrib.frontend_forms.models.FormFieldMixin,
                "djangocms_frontend.frontenduiitem",
            ),
        ),
        migrations.CreateModel(
            name="FormEntry",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("form_name", models.SlugField(verbose_name="Form")),
                (
                    "entry_data",
                    models.JSONField(
                        blank=True,
                        default=dict,
                        encoder=django.core.serializers.json.DjangoJSONEncoder,
                    ),
                ),
                ("html_headers", models.JSONField(blank=True, default=dict)),
                ("entry_created_at", models.DateTimeField(auto_now_add=True)),
                ("entry_updated_at", models.DateTimeField(auto_now=True)),
                (
                    "form_user",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="User",
                    ),
                ),
            ],
            options={
                "verbose_name": "Form entry",
                "verbose_name_plural": "Form entries",
            },
        ),
        migrations.CreateModel(
            name="CaptchaField",
            fields=[],
            options={
                "verbose_name": "reCaptcha field",
                "proxy": True,
                "indexes": [],
                "constraints": [],
            },
            bases=(
                djangocms_frontend.contrib.frontend_forms.models.FormFieldMixin,
                "djangocms_frontend.frontenduiitem",
            ),
        ),
        migrations.CreateModel(
            name="Choice",
            fields=[],
            options={
                "verbose_name": "Choice",
                "proxy": True,
                "indexes": [],
                "constraints": [],
            },
            bases=(
                djangocms_frontend.contrib.frontend_forms.models.FormFieldMixin,
                "djangocms_frontend.frontenduiitem",
            ),
        ),
        migrations.CreateModel(
            name="DateField",
            fields=[],
            options={
                "verbose_name": "Date field",
                "proxy": True,
                "indexes": [],
                "constraints": [],
            },
            bases=(
                djangocms_frontend.contrib.frontend_forms.models.FormFieldMixin,
                "djangocms_frontend.frontenduiitem",
            ),
        ),
        migrations.CreateModel(
            name="DateTimeField",
            fields=[],
            options={
                "verbose_name": "Date field",
                "proxy": True,
                "indexes": [],
                "constraints": [],
            },
            bases=(
                djangocms_frontend.contrib.frontend_forms.models.FormFieldMixin,
                "djangocms_frontend.frontenduiitem",
            ),
        ),
        migrations.CreateModel(
            name="DecimalField",
            fields=[],
            options={
                "verbose_name": "Decimal field",
                "proxy": True,
                "indexes": [],
                "constraints": [],
            },
            bases=(
                djangocms_frontend.contrib.frontend_forms.models.FormFieldMixin,
                "djangocms_frontend.frontenduiitem",
            ),
        ),
        migrations.CreateModel(
            name="EmailField",
            fields=[],
            options={
                "verbose_name": "Email field",
                "proxy": True,
                "indexes": [],
                "constraints": [],
            },
            bases=(
                djangocms_frontend.contrib.frontend_forms.models.FormFieldMixin,
                "djangocms_frontend.frontenduiitem",
            ),
        ),
        migrations.CreateModel(
            name="IntegerField",
            fields=[],
            options={
                "verbose_name": "Integer field",
                "proxy": True,
                "indexes": [],
                "constraints": [],
            },
            bases=(
                djangocms_frontend.contrib.frontend_forms.models.FormFieldMixin,
                "djangocms_frontend.frontenduiitem",
            ),
        ),
        migrations.CreateModel(
            name="TimeField",
            fields=[],
            options={
                "verbose_name": "Date field",
                "proxy": True,
                "indexes": [],
                "constraints": [],
            },
            bases=(
                djangocms_frontend.contrib.frontend_forms.models.FormFieldMixin,
                "djangocms_frontend.frontenduiitem",
            ),
        ),
        migrations.CreateModel(
            name="UrlField",
            fields=[],
            options={
                "verbose_name": "URL field",
                "proxy": True,
                "indexes": [],
                "constraints": [],
            },
            bases=(
                djangocms_frontend.contrib.frontend_forms.models.FormFieldMixin,
                "djangocms_frontend.frontenduiitem",
            ),
        ),
    ]



from django.test import TestCase

from djangocms_frontend.contrib.frontend_forms.forms import FormsForm
from djangocms_frontend.contrib.frontend_forms.models import Form, FormEntry


class FormsModelTestCase(TestCase):
    def test_form_instance(self):
        instance = Form.objects.create()
        instance.initialize_from_form(FormsForm)
        self.assertEqual(str(instance), "Form (1)")
        self.assertEqual(instance.get_short_description(), "<unnamed>")
        instance.config["form_name"] = "my-test-form"
        self.assertEqual(instance.get_short_description(), "(my-test-form)")

        entry = FormEntry.objects.create(form_name=instance.form_name)
        self.assertEqual(str(entry), "my-test-form (1)")



from djangocms_frontend import settings

global_options = settings.FORM_OPTIONS


def get_option(form, option, default=None):
    form_options = getattr(getattr(form, "Meta", None), "options", {})
    return form_options.get(option, global_options.get(option, default))



import hashlib

from django.core.exceptions import ImproperlyConfigured
from django.core.mail import mail_admins, send_mail
from django.template.loader import render_to_string
from django.utils.html import strip_tags
from django.utils.translation import gettext_lazy as _

from .entry_model import FormEntry
from .helper import get_option

_action_registry = {}


def get_registered_actions():
    """Creates a tuple for a ChoiceField to select form"""
    result = tuple(
        (hash, action_class.verbose_name)
        for hash, action_class in _action_registry.items()
    )
    return result if result else ((_("No actions registered"), ()),)


def register(action_class):
    """Function to call or decorator for an Action class to make it available for the
    djangocms_frontend.contrib.frontend_forms plugin"""

    if not issubclass(action_class, FormAction):
        raise ImproperlyConfigured(
            "djangocms_frontend.contrib.frontend_forms.actions.register only "
            "accets subclasses of djangocms_frontend.contrib.frontend_forms.actions.FormAction"
        )
    if not action_class.verbose_name:
        raise ImproperlyConfigured(
            "FormActions need to have a verbose_name property to be " "resitered"
        )
    hash = hashlib.sha1(action_class.__name__.encode("utf-8")).hexdigest()
    _action_registry.update({hash: action_class})
    return action_class


def get_action_class(action):
    return _action_registry.get(action, None)


class FormAction:
    verbose_name = None

    def execute(self, form, request):
        raise NotImplementedError()


@register
class SaveToDBAction(FormAction):
    verbose_name = _("Save form submission")

    def execute(self, form, request):
        if get_option(form, "unique", False) and get_option(
            form, "login_required", False
        ):
            keys = {
                "form_name": get_option(form, "form_name"),
                "form_user": request.user,
            }
            defaults = {}
        else:
            keys = {}
            defaults = {
                "form_name": get_option(form, "form_name"),
                "form_user": request.user,
            }
        defaults.update(
            {
                "entry_data": form.cleaned_data,
                "html_headers": dict(
                    user_agent=request.headers["User-Agent"],
                    referer=request.headers["Referer"],
                ),
            }
        )
        if keys:  # update_or_create only works if at least one key is given
            try:
                FormEntry.objects.update_or_create(**keys, defaults=defaults)
            except FormEntry.MultipleObjectsReturned:  # Delete outdated objects
                FormEntry.objects.filter(**keys).delete()
                FormEntry.objects.create(**keys, **defaults)
        else:
            FormEntry.objects.create(**defaults), True


SAVE_TO_DB_ACTION = next(iter(_action_registry)) if _action_registry else None


@register
class SendMailAction(FormAction):
    verbose_name = _("Send email to administrators")
    recipients = None
    from_mail = None
    template = "djangocms_frontend/actions/mail.html"
    subject = _("%(form_name)s form submission")

    def execute(self, form, request):
        context = dict(
            cleaned_data=form.cleaned_data,
            user=request.user,
            user_agent=request.headers["User-Agent"],
            referer=request.headers["Referer"],
        )
        html_message = render_to_string(self.template, context)
        message = strip_tags(html_message)
        if self.recipients is None:
            mail_admins(
                self.subject % dict(form_name=""),
                message,
                fail_silently=True,
                html_message=html_message,
            )
        else:
            send_mail(
                self.subject % dict(form_name=""),
                message,
                self.recipients,
                self.from_mail,
                fail_silently=True,
                html_message=html_message,
            )



from djangocms_frontend import settings

if settings.framework == "bootstrap5":
    from .bootstrap5.responsive import ResponsiveFormMixin, ResponsiveMixin
else:

    class ResponsiveMixin:
        pass

    class ResponsiveFormMixin:
        pass
